/*
Copyright 2008, Freie Universitaet Berlin. All rights reserved.

These sources implementing a version of the Distance Source Routing 
protocol (DSR)were developed at the Freie Universitt Berlin, for the
subject "Telematic Project: Wireless Embedded Systems".

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the authors Joaquin Gonzalez and
Inas Sameer at the subject "Telematic Project: Wireless Embedded Systems"
during the Winter term 2008.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, December 2008
*/

#ifndef DSR_H_
#define DSR_H_

#include <ScatterWeb.System.h>
#include <ScatterWeb.CC1020.h>
#include <string.h>
#include <stdio.h>
#include "ScatterWeb.Process.h"
#include "transport_layer.h"


/**
 @brief	Minimum node address available in local network.
 */
#define MIN_ADDR 31

/**
 @brief	Maximum node address available in local network.
 */
#define MAX_ADDR 40

/**
 @brief	Network diameter.
 */
#define Net_diam 10

/**
 @brief	Number Data Cache entries.
 */
#define Data_entries 20

/**
 @brief	Maximum number of Route Request retransmissions.
 */
#define RTODSR 3

/**
 @brief	Payload size.
 */
#define PayloadSize 30+40

/**
 * @brief	Predefined local broadcast address
 */
#define LOCAL_BROADCAST MAX_ADDR

/**
 * @brief	DSR Route Request Packet Type
 */
#define DSR_REQUEST_PACKET USERDEFINED_PACKET+1

/**
 * @brief	DSR Route Response Packet Type
 */
#define DSR_RESPONSE_PACKET USERDEFINED_PACKET+2

/**
 * @brief	DSR Error Packet Type
 */
#define DSR_ERROR_PACKET USERDEFINED_PACKET+3

/**
 * @brief	DSR Data Packet Type
 */
#define DSR_DATA_PACKET USERDEFINED_PACKET+4

/**
 @brief	Route Record definition.
 */
typedef netaddr_t Route_record[Net_diam];

/**
 @brief	Payload data definition.
 */
typedef uint8_t payload_data[PayloadSize];

/**
 @brief	Request packet definition.
 */
typedef struct{
	netaddr_t dest; ///< Destination Node Address
	Route_record Route_rec; ///< Route Record
	int pointer; ///< Pointer
	int TTL; ///< Time To Live
}req_packet_t;

/**
 @brief	Response packet definition.
 */
typedef struct{
	netaddr_t source; ///< Source Node Address
	Route_record Route; ///< Route Record
	int TTL; ///< Time To Live
}rsp_packet_t;

/**
 @brief	Data packet definition.
 */
typedef struct{
	netaddr_t dest; ///< Destination Node Address
	Route_record Route; ///< Route Record
	int TTL; ///< Time To Live
	payload_data data; ///< Payload Data
}data_packet_t;

/**
 @brief	Data Cache entry definition.
 */
typedef struct{
	netaddr_t dst; ///< Destination Node Address
	payload_data data; ///< Payload Data
	int reqcount; ///< Route Discovery counter
}data_entry_t;

/**
 @brief	Data Cache type definition.
 */
typedef data_entry_t Data_Cache_t [Data_entries];


/**
 @brief				Data Cache (Send Buffer) as global variable
 */
Data_Cache_t Data_Cache;

/**
 @brief	Request Table entry definition.
 */
typedef struct{
	netaddr_t dst; ///< Destination Node Address
	int flag; ///< Activated Request flag
}req_table_entry_t;

/**
 @brief	Request Table type definition.
 */
typedef req_table_entry_t Req_table_t [MAX_ADDR-MIN_ADDR];

/**
 @brief				Request Table as global variable
 */
Req_table_t Req_table;


//DSR FUNCTION DEFINITIONS

/**
 @brief				function to start Route Discovery Process
 @param[in]			node			Destination node address
 @return					<code>True</code> if successful;
							<code>False</code> otherwise.
 */
bool Start_Route_Discovery(netaddr_t node);

/**
 @brief				function to restart Route Discovery Process
 @param[in]			node			Destination node address
 */
void Restart_Route_Disc (void *datap);

/**
 @brief				function to compare two node addresses
 @param[in]			node1			First node address
 @param[in]			node2			Second node address
 @return					<code>True</code> if equal addresses;
							<code>False</code> otherwise.
 */
bool compAddr(netaddr_t node1, netaddr_t node2);

/**
 @brief				function to check if node address belongs to defined local network address range
 @param[in]			node			Node address
 @return					<code>True</code> if node belongs to local network address range;
							<code>False</code> otherwise.
 */
bool IsNtwAddr(netaddr_t node);

/**
 @brief				function to add node address to ROUTE REQUEST Route Record
 @param[in]			req_pckt		Pointer to REQ packet
 @param[in]			node			Node address
 @return					<code>True</code> if successful;
							<code>False</code> otherwise.
 */
bool Add_Hop_Req_Pckt(req_packet_t* req_pckt, netaddr_t node);

/**
 @brief				function to obtain next hop to a given node in a RSP packet Route Record
 @param[in]			Route		Route Record
 @param[in]			node		Node address
 @return			node address of the next hop in Route.
 */
netaddr_t NextDsrRspHop(Route_record Route, netaddr_t node);

/**
 @brief				function to obtain next hop to a given node in a DATA packet Route Record
 @param[in]			Route		Route Record
 @param[in]			node		Node address
 @return			node address of the next hop in Route.
 */
netaddr_t NextDsrDataHop(Route_record Route, netaddr_t node);

/**
 @brief				function to determine if a node address is contained in a Route Record
 @param[in]			node		Node address
 @param[in]			Route		Route Record
 @param[in]			size		number of entries in Route Record
 @return			<code>True</code> if node address contained in Route Record;
					<code>False</code> otherwise.
 */
bool isAddrInRoute(netaddr_t node,Route_record Route, int size);

/**
 @brief				function to determine node position in Route Record sequence
 @param[in]			Route1		Route Record
 @param[in]			node		Node address
 @return			node position in Route Record.
 */
int Look_up_Position(Route_record Route1, netaddr_t node);

/**
 @brief				function to delete all Data Buffer entries refered to the provided target node
 @param[in]			node		Destination node address
 */
void Delete_Data_Entries(netaddr_t dstnode);

/**
 @brief				function to delete payload data
 @param[in]			data		Payload data
 */
void Clear_Payload(payload_data data);


//Packet initialization functions

/**
 @brief				function to initialize a Request packet
 @param[in]			req_pckt		Pointer to REQ packet
 */
void Init_Req_Pckt(req_packet_t* req_pckt);

/**
 @brief				function to initialize a Response packet
 @param[in]			rsp_pckt		Pointer to RSP packet
 */
void Init_Rsp_Pckt(rsp_packet_t* rsp_pckt);

/**
 @brief				function to initialize a Data packet
 @param[in]			data_pckt		Pointer to DATA packet
 */
void Init_Data_Pckt(data_packet_t* data_pckt);


//Handle packet functions

/**
 @brief				Request packet handling function
 @param[in]			req_pckt		Pointer to REQ packet
 */
void Handle_Req_Pckt(req_packet_t* req_pckt);

/**
 @brief				Response packet handling function
 @param[in]			rsp_pckt		Pointer to RSP packet
 */
void Handle_Rsp_Pckt(rsp_packet_t* rsp_pckt);

/**
 @brief				Data packet handling function
 @param[in]			data_pckt		Pointer to DATA packet
 */
void Handle_Data_Pckt(data_packet_t* data_pckt);


//Send packet functions

/**
 @brief				Function to send Response packet from initiator back to next hop
 @param[in]			Route		Route Record
 @param[in]			nextnode		Next hop node address
 @return			<code>True</code> if packet was successfully buffered;
 					<code>False</code> otherwise.
 */
bool Send_Rsp_Pckt(Route_record Route, netaddr_t nextnode);

/**
 @brief				Function to send all data packets in the send buffer refering to a given target node
 @param[in]			Route		Route Record
 @param[in]			nextnode	Next hop node address
 @param[in]			dstnode		Destination node address
 */
void Send_Data_to_Dest(Route_record Route, netaddr_t nextnode, netaddr_t dstnode);

/**
 @brief				Function to send Response packet from initiator to next hop in the Route Record
 @param[in]			Route		Route Record
 @param[in]			nextnode		Next hop node address
 @param[in]			dstnode		Destination node address
 @return			<code>True</code> if packet was successfully buffered;
 					<code>False</code> otherwise.
 */
bool Send_Data_Pckt(Route_record Route, netaddr_t nextnode, netaddr_t dstnode);


//Forward packet functions

/**
 @brief				Function to forward Request packet to next hop in message's Route Record
 @param[in]			req_pckt		Pointer to REQ packet
 @return			<code>True</code> if REQ packet was successfully buffered;
 *					<code>False</code> otherwise.
 */
bool Forward_Req_Pckt(req_packet_t* req_pckt);

/**
 @brief				Function to forward Response packet to next hop in message's Route Record
 @param[in]			rsp_pckt		Pointer to RSP packet
 @param[in]			nextnode		Next hop node address
 @return			<code>True</code> if RSP packet was successfully buffered;
 *					<code>False</code> otherwise.
 */
bool Forward_Rsp_Pckt(rsp_packet_t* rsp_pckt, netaddr_t nextnode);

/**
 @brief				Function to forward Data packet to next hop in message's Route Record
 @param[in]			data_pckt		Pointer to DATA packet
 @param[in]			nextnode		Next hop node address
 @return			<code>True</code> if DATA packet was successfully buffered;
 *					<code>False</code> otherwise.
 */
bool Forward_Data_Pckt(data_packet_t* data_pckt, netaddr_t nextnode);


//Show DSR packet functions

/**
 @brief				Function to show the content of a Request packet
 @param[in]			req_pckt		Pointer to REQ packet
 */
void Show_Req_Pckt(req_packet_t* req_pckt);

/**
 @brief				Function to show the content of a Response packet
 @param[in]			rsp_pckt		Pointer to RSP packet
 */
void Show_Rsp_Pckt(rsp_packet_t* rsp_pckt);

/**
 @brief				Function to show the content of a Data packet
 @param[in]			rsp_pckt		Pointer to DATA packet
 */
void Show_Data_Pckt(data_packet_t* data_pckt);


//Data cache handling functions

/**
 @brief				Function to initialize the Data Cache
 */
void InitDataCache();

/**
 @brief				Function to insert a new entry in the Data Cache
 @param[in]			dst_node	Destination node address
 @param[in]			data		Payload data
 @return			Pointer to inserted data entry
 */
data_entry_t* Insert_Data_Entry(netaddr_t dstnode, payload_data data);

/**
 @brief				Function to obtain an entry to a given target node in the Data Cache
 @param[in]			dst_node		Destination node address
 @param[in]			data_entry		Pointer to data entry where to copy content
 @return			<code>True</code> if entry to destination was found in the Data Cache;
 *					<code>False</code> otherwise.
 */
bool Get_Data_Entry(netaddr_t dstnode, data_entry_t* data_entry);

/**
 @brief				Function to show the content of the Data Cache
 */
void Show_Data_Cache();

/**
 @brief				Function to initialize Request Table
 */
void InitReqTable();

/**
 @brief				Function to check if a given target node is in Request Table
 @param[in]			dst_node		Destination node address
 @return			<code>True</code> if the given target node is in Request Table;
 *					<code>False</code> otherwise.
 */
bool isAddrInReqTable(netaddr_t dstnode);

/**
 @brief				Function to insert an entry to a given target node in Request Table
 @param[in]			dst_node		Destination node address
 @return			<code>True</code> if the entry was correctly created in Request Table;
 *					<code>False</code> otherwise.
 */
bool Insert_ReqTable_Entry(netaddr_t dstnode);

/**
 @brief				Function to delete a given target node from Request Table
 @param[in]			dst_node		Destination node address
 @return			<code>True</code> if the entry was correctly deleted in Request Table;
 *					<code>False</code> otherwise.
 */
bool Delete_ReqTable_Entry(netaddr_t dstnode);

/**
 @brief				Function to show Request Table
 */
void Show_Req_Table();

/**
 @brief				Function to send transport header and payload data using the DSR protocol
 @param[in]			dst_node		Destination node address
 @param[in]			
 @return			<code>True</code> if the entry was correctly deleted in Request Table;
 *					<code>False</code> otherwise.
 */
bool SendDSR(netaddr_t dstnode, payload_data data);


#endif /*DSR_H_*/
