/*
Copyright 2008, Freie Universitaet Berlin. All rights reserved.

These sources implementing a version of the layer 4 Transmission
Control Protocol (TCP) were developed at the Freie Universitt Berlin, 
for the subject "Telematic Project: Wireless Embedded Systems".

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the authors Joaquin Gonzalez and
Inas Sameer at the subject "Telematic Project: Wireless Embedded Systems"
during the Winter term 2008.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, December 2008
*/

#include <string.h>
#include <ScatterWeb.System.h>
#include <ScatterWeb.CC1020.h>
#include "transport_layer.h"
#include "dsr.h"


//IMPLEMENTATION OF TRANSPORT PROTOCOL FUNCTIONS

//Function to initialize the TCP protocol
void InitTCP(){

	InitSendBuffer();
	InitTCBList();

}

//Function to initialize the TCB list
void InitTCBList(){

	int i;
	for (i=0; i<TCBListSize; i++){
		TCBList[i].dst = 0;
		TCBList[i].TCP_state = ST_LISTEN;
		TCBList[i].nextACK_number = 0;
		TCBList[i].nextSEQ_number = 0;
		TCBList[i].TxWindow = TAM_TXWNDW;
		TCBList[i].reconncount = 0;
		TCBList[i].nextSegment = 0;
		TCBList[i].sgmtcount = 0;
	}
}

//Function to insert new connection in the TCB list
bool CreateTCBEntry(netaddr_t dstnode){

	int i;
	bool found = false;
	for (i=0; i<TCBListSize; i++){
		if ((TCBList[i].dst == 0) && (!found)){
			TCBList[i].dst = dstnode;
			TCBList[i].nextACK_number = 1;
			TCBList[i].nextSEQ_number = 1;
			TCBList[i].TxWindow = 0;
			TCBList[i].TCP_state = SYN_SENT;
			TCBList[i].reconncount = 0;
			TCBList[i].nextSegment = 0;
			TCBList[i].sgmtcount = 0;
			found = true;
		}
	}
	return found;
}

//Function to determine TCB position in TCB List
int Look_up_TCB_Position(netaddr_t node){

	int i, pos;
	pos = TCBListSize;
	for(i=0; i<Net_diam; i++){
		if (compAddr(TCBList[i].dst, node)) pos=i;
	}
	return pos;
}

//Function to check if there is a TCB for the given destination node
bool isConnectionInTCBList(netaddr_t dstnode){

	int i;
	bool found = false;
	for (i=0; i<TCBListSize; i++){
		if (TCBList[i].dst == dstnode) found = true;
	}
	return found;
}


//Function to initialize the Send Buffer
void InitSendBuffer(){
	int i;
	for (i=0; i<(TAM_BUFF); i++){
		SendBuffer[i].dst = 0;
		SendBuffer[i].SEQ_number = 0;
		SendBuffer[i].retxcount = 0;
		SendBuffer[i].sgmnmbr = 0;
		Clear_TCP_Payload(SendBuffer[i].data);
	}
}

//Function to delete TCP Payload data
void Clear_TCP_Payload(tcp_payload data){
	int i;
	for(i=0; i<TCPPayloadSize; i++){
		data[i] = 0;
	}
}

//Transport packet handling function
void Handle_Tcp_Data_Pckt(tcp_packet_t* tcp_pckt, uint8_t sndnode){

	if(DEBUGTCP) {
		printf("Received a TCP packet from %d! \n", sndnode);
		Show_TCP_Packet(tcp_pckt);
	}

	//check out if receiver has a TCB entry for the corresponding sender node

	if (!isConnectionInTCBList(sndnode)){
		printf("No TCB existent for sender node. Checking if segment is connection request...\n");

		//create new TCB entry for given node
		if (!CreateTCBEntry(sndnode)){
			printf("ERROR: no space available in TCB List for new entry to %d!", sndnode);
		}
		else{//new TCB entry created
			//check if it is a connection request segment
			if (isSYNsegment(tcp_pckt)){
				printf("CONNECTION REQUEST received from %d\n", sndnode);
				if (DEBUGTCP) Show_TCP_Packet(tcp_pckt);

				//update newly created TCB entry to SYN_RCVD state 
				setTCBState(sndnode, SYN_RCVD);

				//set next ACK number
				setTCBACK(sndnode,tcp_pckt->SEQ_number+1);
				if (DEBUGTCP) ShowTCBList();

				//send connection response (SYN-ACK) back to sender
				if (SendTCPConnResp(sndnode)) printf("Connection response sent back to %d\n", sndnode);

				//increment SEQ number for the connection
				updateTCBSEQ(sndnode);
			}
			else{
				printf("Segment received from sender without established connection. Discard...\n");
			}
		}
	}
	else{//existent TCB entry for current segment sender node

		if (DEBUGTCP) printf("Current TCP connection state with node %d: %d\n", sndnode, getTCBState(sndnode));

		//check current TCP connection state
		switch (getTCBState(sndnode))
		{
			//TCP state ST_LISTEN: the server is waiting for an incoming call.
			case ST_LISTEN:

				break;

			//TCP state SYN_SENT: the application has started to open a connection.
			case SYN_SENT:
	
				//check if it is a connection request segment
				if (isSYNACKsegment(tcp_pckt)){
					printf("CONNECTION RESPONSE received from %d\n", sndnode);
					if (DEBUGTCP) Show_TCP_Packet(tcp_pckt);

					//update TCB entry to ESTAB state
					setTCBState(sndnode, ESTAB);
					//set next ACK number
					setTCBACK(sndnode,tcp_pckt->SEQ_number+1);
					//increment Tx window size for the connection
					setTCBTxWndw(sndnode,getTCBTxWndw(sndnode)+1);

					//send connection confirmation (ACK of SYN) to receiver
					if (SendTCPConnConf(sndnode)) printf("Connection confirmation sent back to %d\n", sndnode);

					//Remove Connection Establishment Timer
					Timers_remove(Restart_Three_Way_Handshake,sndnode);

					printf("CONNECTION ESTABLISHED with %d!\n", sndnode);
					P5OUT ^= 0x80;
					if (DEBUGTCP) ShowTCBList();
				}
				else{
					if(isSYNsegment(tcp_pckt)){//simultaneous open connection
						printf("\nRECEIVED SIMULTANEOUS CONNECTION REQUEST FROM %d\n",sndnode);
						Show_TCP_Packet(tcp_pckt);
						
						//update ACK number (last SEQ segment correctly delivered + 1)
						setTCBACK(sndnode,tcp_pckt->SEQ_number+1);

						//send Acknowledgement segment back to sender
						SendACK(sndnode);	

						//update TCB entry to SYN_RCVD state
						setTCBState(sndnode, SYN_RCVD);

					}
					else{
						printf("Segment discarded in connection establishment.\n");
					}
				}
				break;

			//TCP state SYN_RCVD: a connection request has arrived, wait for ACK.
			case SYN_RCVD:

				//check if received segment is a connection confirmation (ACK of SYN) and has a valid SEQ number
				if(isValidSEQ(sndnode, tcp_pckt->SEQ_number) && isACKactive(tcp_pckt) && isnextACK(sndnode,tcp_pckt->ACK_number)){
					printf("CONNECTION CONFIRMATION from %d\n", sndnode);
					if (DEBUGTCP) Show_TCP_Packet(tcp_pckt);

					//update TCB entry to ESTAB state
					setTCBState(sndnode, ESTAB);
					
					//increment Tx window size for the connection
					setTCBTxWndw(sndnode,getTCBTxWndw(sndnode)+1);

					//we don't increment ACK number for the connection
						//the received connection confirmation is an empty segment containing an ACK 
						//that does not occupy sequence number space
						//otherwise we would wind up ACKing ACK's
	
					printf("CONNECTION ESTABLISHED with %d!\n", sndnode);
					P5OUT ^= 0x80;
					if (DEBUGTCP) ShowTCBList();
						
				}
				else{//invalid SEQ number
					printf("Received segment out of reception window. Segment discarded.\n");
				}
				break;

			//TCP state ESTAB: opened connection, normal data transfer.
			case ESTAB:

				//check if received segment has a valid SEQ number
				if(!isValidSEQ(sndnode, tcp_pckt->SEQ_number)){

					//check if the segment contains data
					if (isPSHactive(tcp_pckt)){
						printf("\nRECEIVED DUPLICATED DATA SEGMENT SEQ %d FROM %d\n",tcp_pckt->SEQ_number,sndnode);

						//send Acknowledgement segment back to sender
						SendACK(sndnode);
					}
					printf("Received segment out of reception window. Segment discarded.\n");
				}
				else{//valid segment SEQ number

					//check if the segment acknowledges confirmation of previously sent data segment
					if (isACKactive(tcp_pckt)){
						if (isnextACK(sndnode,tcp_pckt->ACK_number)){
							printf("ACK received for segment SEQ %d!\n", tcp_pckt->ACK_number-1);
							//increment Tx window size for the connection
							setTCBTxWndw(sndnode,getTCBTxWndw(sndnode)+1);

							//remove timer using first Send Buffer entry
							Timers_remove( Restart_SendTCP,getSendBufferEntry(sndnode,tcp_pckt->ACK_number-1));

							//delete Send Buffer entry
							deleteSendBufferEntry(sndnode,tcp_pckt->ACK_number-1);

							//check out if there is a segment waiting to be sent at Send Buffer
							if(LookupWaitingSegment(sndnode)){
								sendbuffer_entry_t* entry;
								entry = getNextBuffSegm(sndnode);
								//set segment SEQ number
								entry->SEQ_number = getTCBSEQ(sndnode);

								//send packet
								if (SendTCP(entry)) printf("Segment %d: %s sent to node %d\n",SendDataCtr,entry->data,sndnode);

								//increment next segment number
								incTCBSegmNumbr(sndnode);

								//increment number of sent TCP segments
								SentTCPDataCtr++;

								//increment SEQ number for the connection
								updateTCBSEQ(sndnode);

								Timers_add(RTOTCP*1024, Restart_SendTCP, entry);

								//decrement Tx window size for the connection
								setTCBTxWndw(sndnode,getTCBTxWndw(sndnode)-1);
								ShowTCBList();
							}
							else{
								printf("No segments waiting to be sent after ACK!\n");
								ShowSendBuffer();
								ShowTCBList();
							}
						}
					}

					//check if the segment contains data
					if (isPSHactive(tcp_pckt)){
						printf("\nRECEIVED DATA SEGMENT %d FROM %d\n",RcvDataPcktCtr,sndnode);
						Show_TCP_Packet(tcp_pckt);

						//increase number of received TCP segments
						RcvTCPDataCtr++;
						
						//update ACK number (last SEQ segment correctly delivered + 1)
						setTCBACK(sndnode,tcp_pckt->SEQ_number+1);

						//send Acknowledgement segment back to sender
						SendACK(sndnode);
					}

					//check if the segment is a close connection request
					if (isFINsegment(tcp_pckt)){

						printf("CONNECTION TERMINATION REQUEST received from %d\n", sndnode);
						//set next ACK number
						setTCBACK(sndnode,tcp_pckt->SEQ_number+1);	

						//send connection response (SYN-ACK) back to sender
						if (SendTCPFINACK(sndnode)) printf("Close connection response sent back to %d\n", sndnode);

						//increment SEQ number for the connection
						updateTCBSEQ(sndnode);

						//update TCB entry to FINWAIT1 state
						setTCBState(sndnode, LAST_ACK);
					}
				}
				break;

			//TCP state FINWAIT1: the application has said the connection is finished.
			case FINWAIT1:

				//check out if it is a close connection response segment
				if (isFINACKsegment(tcp_pckt) && isnextACK(sndnode,tcp_pckt->ACK_number)){

					//remove timer using first Send Buffer entry
					Timers_remove(Restart_Close_Connection, sndnode);

					//set next ACK number
					setTCBACK(sndnode,tcp_pckt->SEQ_number+1);

					//send Acknowledgement segment back to sender
					SendACK(sndnode);

					//update TCB entry to TIME_WAIT state
					setTCBState(sndnode, TIME_WAIT);

					if (DEBUGTCP) printf("Received FIN-ACK segment. Connection set to TIME_WAIT modus\n");
					
					//Set waiting timer to release connection
					Timers_add(CloseTCPTimeout*1024, CloseSenderConnection, sndnode);

				}
				else{ 
					if(isFINsegment(tcp_pckt)){//simultaneous close
						printf("\nRECEIVED SIMULTANEOUS CONNECTION TERMINATION FROM %d\n",sndnode);
						Show_TCP_Packet(tcp_pckt);
						
						//update ACK number (last SEQ segment correctly delivered + 1)
						setTCBACK(sndnode,tcp_pckt->SEQ_number+1);

						//send Acknowledgement segment back to sender
						SendACK(sndnode);	

						//update TCB entry to CLOSING state
						setTCBState(sndnode, CLOSING);
					}
					else{
						printf("Segment discarded");
					}
				}
				break;

			//TCP state CLOSING: CLOSING: both sides have tried to close connection simultaneously.
			case CLOSING:

				//check if the segment acknowledges confirmation of previously sent data segment
				if (isACKactive(tcp_pckt) && isnextACK(sndnode,tcp_pckt->ACK_number)){
					
					//remove timer using first Send Buffer entry
					Timers_remove(Restart_Close_Connection, sndnode);

					//update TCB entry to TIME_WAIT state
					setTCBState(sndnode, TIME_WAIT);

					if (DEBUGTCP) printf("Received ACK for FIN. Connection set to TIME_WAIT modus\n");
					
					//Set waiting timer to release connection
					Timers_add(CloseTCPTimeout*1024, CloseSenderConnection, sndnode);
				}

				break;

			//TCP state TIME_WAIT: wait for all packets to die off.
			case TIME_WAIT:
				break;

			//TCP state LAST_ACK: wait for incoming connection close connection acknowledgement.
			case LAST_ACK:

				//check if received segment is a close connection confirmation (ACK of FIN) and has a valid SEQ number
				if(isValidSEQ(sndnode, tcp_pckt->SEQ_number) && isACKactive(tcp_pckt) && isnextACK(sndnode,tcp_pckt->ACK_number)){

					printf("Received close connection confirmation. Closing connection...\n");
					
					//Remove all active timers to the given target node
					RemoveActiveTimers(sndnode);
					//Erase all Send Buffer entries to given target node
					ClearSendBufferEntries(sndnode);
					//Erase TCB entry for destination node
					DeleteTCBEntry(sndnode);
					P5OUT ^= 0x80;
					
					if (DEBUGTCP) printf("Connection resources liberated to %d\n",sndnode);
					printf("CONNECTION CLOSED to %d\n",sndnode);
					ShowTCPTestReceiver();

				}
				else{//invalid SEQ number
						printf("Received segment with unexpected ACK. Segment discarded.\n");
				}
				break;

			//TCP state CLOSED: no connection is open or pending.
			case CLOSED:
				break;

		}//end of TCP state switch
	}

}


//Function to insert a new entry in the Send Buffer
sendbuffer_entry_t* Insert_Entry_SendBuffer(netaddr_t dstnode, tcp_payload data){

	int i, pos;
	pos = 0;
	bool found = false;

	//traverse Send Buffer to look for a free entry (dst == 0)
	for(i=0; i<TAM_BUFF; i++){
		if((SendBuffer[i].dst == 0) && (!found)){
			SendBuffer[i].dst = dstnode;
			strcpy(SendBuffer[i].data,data);
			SendBuffer[i].SEQ_number = 0;
			SendBuffer[i].retxcount = 0;
			//assign TCB next segment number
			SendBuffer[i].sgmnmbr = getTCBSegmCtr(dstnode);
			//increment TCB next segment number 
			incTCBSegmCtr(dstnode);
			found = true;
		}
	}

	if (!found) {
		return 0;
	}
	else {
		return &SendBuffer[pos];
	}
}

//Function to obtain the TCB connection state to a given target node in TCB List
TCP_state_t getTCBState(netaddr_t dstnode){

	return (TCBList[Look_up_TCB_Position(dstnode)].TCP_state); 
}

//Function that increases the number of connection attempts of a given TCB
void incTCBConnAttempt(netaddr_t dstnode){

	TCBList[Look_up_TCB_Position(dstnode)].reconncount ++;
}

//Function that returns the number of connection attempts of a given TCB
uint8_t getTCBConnAttempt(netaddr_t dstnode){

	return (TCBList[Look_up_TCB_Position(dstnode)].reconncount);
}


//Function that obtains the next segment number to be sent from send buffer for a given TCB
uint8_t getTCBSegmNumbr(netaddr_t dstnode){
	
	return (TCBList[Look_up_TCB_Position(dstnode)].nextSegment);
}

//Function that increments the next segment number to be sent from send buffer for a given TCB
void incTCBSegmNumbr(netaddr_t dstnode){
	TCBList[Look_up_TCB_Position(dstnode)].nextSegment ++;
	if (TCBList[Look_up_TCB_Position(dstnode)].nextSegment >= 512) 
		TCBList[Look_up_TCB_Position(dstnode)].nextSegment = 1;
}

//Function that increments the segment number counter to be sent from send buffer for a given TCB
void incTCBSegmCtr(netaddr_t dstnode){
	TCBList[Look_up_TCB_Position(dstnode)].sgmtcount ++;
	if (TCBList[Look_up_TCB_Position(dstnode)].sgmtcount >= 512) 
		TCBList[Look_up_TCB_Position(dstnode)].sgmtcount = 1;
}

//Function that obtains the segment number counter to be sent from send buffer for a given TCB
uint8_t getTCBSegmCtr(netaddr_t dstnode){
	return (TCBList[Look_up_TCB_Position(dstnode)].sgmtcount);
}


//Function to initialize TCP packet
void Init_TCP_Pckt(tcp_packet_t* tcp_pckt){

	memset(tcp_pckt, 0, sizeof(tcp_packet_t)); // clear struct
	
	//initially disabled control flags
	tcp_pckt->control = 0;
}

//Function to open a TCP connection to a destination node
void OpenTCP(netaddr_t dstnode){

	if (isConnectionInTCBList(dstnode)){
		printf("ERROR: connection already existent to node: %d\n",dstnode);
	}
	else{
		if (!CreateTCBEntry(dstnode)){
			printf("ERROR: no space available in TCB List!");	
		}
		else{

			ShowTCBList();
			if (Start_Three_Way_Handshake(dstnode)) printf("Establishing connection to %d\n",dstnode);

			//increment SEQ number for the connection
			updateTCBSEQ(dstnode);

			//initialize Connection Establishment Timer
			Timers_add(ConnEst_timeout*1024, Restart_Three_Way_Handshake, dstnode);
		}
	}
}

//Function to start Three Way Handshake
bool Start_Three_Way_Handshake(netaddr_t node){

	tcp_packet_t tcp_pckt;

	payload_data dsrdata;

	//encapsulate TCP segment header in data field before sending it
	Init_TCP_Pckt(&tcp_pckt);
	tcp_pckt.SEQ_number = getTCBSEQ(node);
	tcp_pckt.control = SYN_CB;

	ShowTCBList();

	memcpy(dsrdata, (uint8_t*) &tcp_pckt, sizeof(tcp_pckt));
	// queue packet for sending
	//return Net_send( &npsargs );

	Show_TCP_Packet((tcp_packet_t*)dsrdata);

	//send TCP segment using layer 3 (DSR) primitive
	return SendDSR(node, dsrdata);

}

//Function to show the content of a TCP segment
void Show_TCP_Packet(tcp_packet_t* tcp_pckt){

	printf("TCP packet content: \n");
	printf("   SEQ: %d\n",tcp_pckt->SEQ_number);
	printf("   Control byte: %d\n",tcp_pckt->control);
	
	if ((tcp_pckt->control & ACK_CB)== ACK_CB)
		printf("   ACK active: %d\n", tcp_pckt->ACK_number);

	if ((tcp_pckt->control & SYN_CB)==SYN_CB)
		printf("   SYN active\n");

	if ((tcp_pckt->control & PUSH_CB)==PUSH_CB){
		printf("   PSH active\n");
		printf("   Payload: %s",tcp_pckt->data);
	}

	if ((tcp_pckt->control & FIN_CB)==FIN_CB)
		printf("   FIN active\n");

	printf("\n\n");
}

//Function to show Request Table
void ShowTCBList(){
	printf("TCB List content:\n");
	uint8_t i;
	for(i=0; i<TCBListSize; i++){
		if (!compAddr(TCBList[i].dst, 0)){
			printf("Dst: %d State: %d NextSEQ: %d NextACK: %d\n", TCBList[i].dst, TCBList[i].TCP_state, TCBList[i].nextSEQ_number, TCBList[i].nextACK_number);
			printf("        TxWdw: %d NxtSgmt: %d SgmtCtr: %d", TCBList[i].TxWindow,TCBList[i].nextSegment, TCBList[i].sgmtcount);
		}
	}
	printf("\n\n");
}

//Function to show the Retransmission Buffer (Send Buffer) content
void ShowSendBuffer(){

	printf("TCP Send Buffer content:\n");
	uint8_t i;
	for(i=0; i<TAM_BUFF; i++){
		if (!compAddr(SendBuffer[i].dst, 0)){
			printf("Dst: %d SEQ: %d SegmNmr: %d Ini: %d Rtx: %d Payload: %s\n", SendBuffer[i].dst, SendBuffer[i].SEQ_number, SendBuffer[i].sgmnmbr,SendBuffer[i].inicount, SendBuffer[i].retxcount, SendBuffer[i].data);
		}
	}
}

//Function to check if segment is a connection request
bool isSYNsegment(tcp_packet_t* tcp_pckt){

	if((tcp_pckt->control == SYN_CB)){
		return true;
	}
	else return false;
}

//Function to check if segment is a connection response
bool isSYNACKsegment(tcp_packet_t* tcp_pckt){

	if (tcp_pckt->control == SYNACK_CB){
		return true;
	}
	else return false;
}


//Function to check if segment is a close connection request
bool isFINsegment(tcp_packet_t* tcp_pckt){

	if (tcp_pckt->control == FIN_CB){
		return true;
	}
	else{
		return false;
	}
}

//Function to check if segment is a close connection response
bool isFINACKsegment(tcp_packet_t* tcp_pckt){

	if (tcp_pckt->control == FINACK_CB){
		return true;
	}
	else{
		return false;
	}
}


//Function to check if segment has its ACK control bit active
bool isACKactive(tcp_packet_t* tcp_pckt){

	if ((tcp_pckt->control & ACK_CB)==ACK_CB){
		return true;
	}
	else{
		return false;
	}
}

//Function to check if segment has its PSH control bit active (contains data)
bool isPSHactive(tcp_packet_t* tcp_pckt){

	if ((tcp_pckt->control & PUSH_CB)==PUSH_CB){
		return true;
	}
	else{
		return false;
	}
}


//Function to obtain the SEQ number of a given TCB entry
uint8_t getTCBSEQ(netaddr_t dstnode){

	return TCBList[Look_up_TCB_Position(dstnode)].nextSEQ_number;
	
}

//Function to obtain the ACK number of a given TCB entry
uint8_t getTCBACK(netaddr_t dstnode){

	return TCBList[Look_up_TCB_Position(dstnode)].nextACK_number;
	
}

//Function to obtain the transmission window size of a given TCB entry
uint8_t getTCBTxWndw(netaddr_t dstnode){

	return TCBList[Look_up_TCB_Position(dstnode)].TxWindow;

}

//Function to set the Transmission window size value of a given TCB entry
void setTCBTxWndw(netaddr_t dstnode, uint8_t newTxWndw){

	TCBList[Look_up_TCB_Position(dstnode)].TxWindow = newTxWndw;

}

//Function to increase the SEQ number of a TCB entry for a given target node
void updateTCBSEQ(netaddr_t dstnode){

	TCBList[Look_up_TCB_Position(dstnode)].nextSEQ_number++;
	if (TCBList[Look_up_TCB_Position(dstnode)].nextSEQ_number >= 512) 
		TCBList[Look_up_TCB_Position(dstnode)].nextSEQ_number = 1;
		
}

//Function to set the connection state of a target node TCB
void setTCBState(netaddr_t dstnode, TCP_state_t TCBstate){

	TCBList[Look_up_TCB_Position(dstnode)].TCP_state = TCBstate;

}

//Function to set the ACK number of a TCB for a given target node
void setTCBACK(netaddr_t dstnode,uint8_t ACKnumber){

	TCBList[Look_up_TCB_Position(dstnode)].nextACK_number = ACKnumber;

}

//Function to check if a received segment has a valid SEQ number in connection
bool isValidSEQ(netaddr_t dstnode, uint8_t rcvdSEQ){

	if (DEBUGTCP) printf("Received SEQ: %d , Next expected SEQ: %d\n", rcvdSEQ, TCBList[Look_up_TCB_Position(dstnode)].nextACK_number);
	return (TCBList[Look_up_TCB_Position(dstnode)].nextACK_number == rcvdSEQ);

}

//Function to check if a received segment acknowledges the reception of a previously sent segment
bool isnextACK(netaddr_t dstnode, uint8_t rcvdACK){

	return (TCBList[Look_up_TCB_Position(dstnode)].nextSEQ_number == rcvdACK);
}

//Function to send a TCP connection request (SYN) to a given target node
bool SendTCPConnResp(netaddr_t dstnode){

	tcp_packet_t tcp_pckt;

	//encapsulate TCP segment header in data field before sending it
	Init_TCP_Pckt(&tcp_pckt);
	tcp_pckt.ACK_number = getTCBACK(dstnode);
	tcp_pckt.SEQ_number = getTCBSEQ(dstnode);
	//set TCP control flags
	tcp_pckt.control = SYNACK_CB;

	payload_data dsrdata;

	memcpy(dsrdata, (uint8_t*) &tcp_pckt, sizeof(tcp_pckt));

	//send TCP segment using layer 3 (DSR) primitive
	return SendDSR(dstnode, dsrdata);

}

//Function to send a TCP connection confirmation (SYN-ACK) to a given target node
bool SendTCPConnConf(netaddr_t dstnode){

	tcp_packet_t tcp_pckt;

	//encapsulate TCP segment header in data field before sending it
	Init_TCP_Pckt(&tcp_pckt);
	tcp_pckt.ACK_number = getTCBACK(dstnode);
	tcp_pckt.SEQ_number = getTCBSEQ(dstnode);
	tcp_pckt.control = ACK_CB;
	
	//we don't increment SEQ number for the connection
		//the ACK does not occupy sequence number space
		//otherwise we would wind up ACKing ACK's

	ShowTCBList();

	payload_data dsrdata;

	memcpy(dsrdata, (uint8_t*) &tcp_pckt, sizeof(tcp_pckt));

	Show_TCP_Packet((tcp_packet_t*)dsrdata);

	//send TCP segment using layer 3 (DSR) primitive
	return SendDSR(dstnode, dsrdata);
}

//Function to send a TCP data segment to a given target node
bool SendTCP(void *datap){

	sendbuffer_entry_t* e = (sendbuffer_entry_t*) datap;

	tcp_packet_t tcp_pckt;

	//encapsulate TCP segment header in data field before sending it
	Init_TCP_Pckt(&tcp_pckt);
	tcp_pckt.ACK_number = getTCBACK(e->dst);
	tcp_pckt.SEQ_number = e->SEQ_number;

	tcp_pckt.control = PUSHACK_CB;
	strcpy(tcp_pckt.data,e->data);

	payload_data dsrdata;

	memcpy(dsrdata, (uint8_t*) &tcp_pckt, sizeof(tcp_pckt));

	Show_TCP_Packet((tcp_packet_t*)dsrdata);

	//send TCP segment using layer 3 (DSR) primitive
	return SendDSR(e->dst, dsrdata);
}


//Function to restart Three Way Handshake for a given destination node
void Restart_Three_Way_Handshake (void *datap){

	uint8_t dstnode = (uint8_t) datap;

	printf("Retrying Three-Way Handshake to node %d ... \n",dstnode);

	//increment number of connection attempts in TCB
	incTCBConnAttempt(dstnode);

	//check if maximum number of retransmissions reached
	if (getTCBConnAttempt(dstnode) > ConnEstLimit){
		printf("Maximum re-sending times reached. Terminating connection to dst node %d...\n",dstnode);
		//Erase TCB entry for destination node
		DeleteTCBEntry(dstnode);
		printf("Deleted TCB entry to %d\n",dstnode);
		
	}else{
		//Restart Three Way Handshake to given destination node
		if (Start_Three_Way_Handshake(dstnode)) printf("Three Way Handshake %d sent to node %d\n",getTCBConnAttempt(dstnode),dstnode);
		//Call Restart Three Way Handshake Timer
		Timers_add(ConnEst_timeout*1024, Restart_Three_Way_Handshake, dstnode);
	}
}


//Function to restart Data Segment Transmission
void Restart_SendTCP (void *datap){

	sendbuffer_entry_t* entry = (sendbuffer_entry_t*) datap;

	printf("Data content: %s\n",entry->data);

	entry->retxcount++;

	//check if maximum number of retransmissions reached
	if (entry->retxcount > 3){
		printf("Maximum re-sending times reached. Erasing Send Buffer entry to dst node %d...\n",entry->dst);
		//erase data entry from send buffer
		entry->dst = 0;
		entry->inicount = 0;
		entry->retxcount = 0;
		entry->SEQ_number = 0;
		Clear_TCP_Payload(entry->data);

		ShowSendBuffer();
		
	}else{
		if (SendTCP(entry)) printf("Segment SEQ %d retransmitted to node %d\n",entry->SEQ_number,entry->dst);
		Timers_add(RTOTCP*1024, Restart_SendTCP, entry);

		//increment number of data segments retransmitted from layer 4 protocol
		RetxTCPDataCtr++;
	}
}

//Function to retransmit a TCP data segment
void RetryTCPSend (void *datap){

	sendbuffer_entry_t* entry = (sendbuffer_entry_t*) datap;

	entry->inicount ++;

	//check if maximum number of TCP data retransmissions reached
	if (entry->inicount > DataRetxLimit){
		printf("Maximum initial send attempts reached. Closing connection to dst node %d...\n",entry->dst);
		
		//Liberate connection resources to destination node
		CloseSenderConnection(entry->dst);
		ShowSendBuffer();
		
	}else{
	
		//check if connection has been established
		if((getTCBState(entry->dst) == ESTAB)&&(getTCBTxWndw(entry->dst)>0)){
			entry->SEQ_number = getTCBSEQ(entry->dst);
			if (SendTCP(entry)) printf("Segment %d sent to node %d\n",SentDataPcktCtr,entry->dst);

			//increment number of sent TCP segments
			SentTCPDataCtr++;
			//increment SEQ number for the connection
			updateTCBSEQ(entry->dst);
			//decrement Tx window size for the connection
			setTCBTxWndw(entry->dst,getTCBTxWndw(entry->dst)-1);
			ShowTCBList();

			//initialize retransmission timer
			Timers_add(RTOTCP*1024, Restart_SendTCP, entry);
		}
		else{
			printf("Connection request to %d in progress...\n",entry->dst);
			ShowSendBuffer();
			Timers_add(RetryTCPSend_timeout*1024, RetryTCPSend, entry);
		}
	}
}

//Function to send an Acknowledgement segment to a given target node
bool SendACK(netaddr_t dstnode){

	tcp_packet_t tcp_pckt;

	//encapsulate TCP segment header in data field before sending it
	Init_TCP_Pckt(&tcp_pckt);
	tcp_pckt.ACK_number = getTCBACK(dstnode);
	tcp_pckt.SEQ_number = getTCBSEQ(dstnode);
	tcp_pckt.control = ACK_CB;
	
	//we don't increment SEQ number for the connection
		//the ACK does not occupy sequence number space
		//otherwise we would wind up ACKing ACK's

	ShowTCBList();

	payload_data dsrdata;
	memcpy(dsrdata, (uint8_t*) &tcp_pckt, sizeof(tcp_pckt));

	//send TCP segment using layer 3 (DSR) primitive
	return SendDSR(dstnode, dsrdata);

}

//Function to obtain a Send Buffer entry to a given target node and SEQ number
sendbuffer_entry_t* getSendBufferEntry(netaddr_t dstnode, uint8_t SEQnum){

	int i, pos;
	pos = 0;
	bool found = false;

	//traverse Send Buffer to look for the corresponding entry
	for(i=0; i<TAM_BUFF; i++){
		if (!found){
			//if entry found, assign position
			if (compAddr(SendBuffer[i].dst, dstnode) && (SendBuffer[i].SEQ_number == SEQnum)){
				pos = i;
				found = true;
			}
		}
	}

	if (!found) {
		return 0;
	}
	else {
		return &SendBuffer[pos];
	}
}

//Function to obtain the next send buffer entry waiting to be sent
sendbuffer_entry_t* getNextBuffSegm(netaddr_t dstnode){

	int i, pos;
	pos = 0;
	bool found = false;

	//traverse Send Buffer to look for the corresponding entry
	for(i=0; i<TAM_BUFF; i++){
		if (!found){
			//if entry found, assign position
			if (compAddr(SendBuffer[i].dst, dstnode) && (SendBuffer[i].sgmnmbr == getTCBSegmNumbr(dstnode))){
				pos = i;
				found = true;
			}
		}
	}

	if (!found) {
		return 0;
	}
	else {
		return &SendBuffer[pos];
	}
}


//Function to delete a Send Buffer entry to a given target node
void deleteSendBufferEntry(netaddr_t dstnode, uint8_t SEQnum){

	int i;
	bool found = false;

	//traverse Send Buffer to look for the corresponding entry
	for(i=0; i<TAM_BUFF; i++){
		if (!found){
			//if entry found, erase its content
			if (compAddr(SendBuffer[i].dst, dstnode) && (SendBuffer[i].SEQ_number == SEQnum)){
				found = true;
				SendBuffer[i].dst = 0;
				SendBuffer[i].inicount = 0;
				SendBuffer[i].retxcount = 0;
				SendBuffer[i].sgmnmbr = 0;
				Clear_TCP_Payload(SendBuffer[i].data);
			}
		}
	}
}

//Function to check if there are waiting segments to a given target node in Send Buffer
bool LookupWaitingSegment(netaddr_t dstnode){

	int i;
	bool found = false;

	//traverse Send Buffer to look for the corresponding entry
	for(i=0; i<TAM_BUFF; i++){
		if (!found){
			//if entry found, assign position
			if (compAddr(SendBuffer[i].dst, dstnode) && (SendBuffer[i].sgmnmbr == getTCBSegmNumbr(dstnode))){
				found = true;
			}
		}
	}
	return found;
}

//Function to delete a TCB entry to a given node
void DeleteTCBEntry(netaddr_t dstnode){

	TCBList[Look_up_TCB_Position(dstnode)].dst = 0;
	TCBList[Look_up_TCB_Position(dstnode)].nextACK_number = 1;
	TCBList[Look_up_TCB_Position(dstnode)].nextSEQ_number = 0;
	TCBList[Look_up_TCB_Position(dstnode)].reconncount = 0;
	TCBList[Look_up_TCB_Position(dstnode)].TCP_state = 0;
	TCBList[Look_up_TCB_Position(dstnode)].TxWindow = TAM_TXWNDW;

}


//Function to init Close Connection procedure to given target node
bool Init_Close_Connection(netaddr_t node){

	tcp_packet_t tcp_pckt;

	//encapsulate TCP segment header in data field before sending it
	Init_TCP_Pckt(&tcp_pckt);
	tcp_pckt.SEQ_number = getTCBSEQ(node);
	tcp_pckt.control = FIN_CB;

	ShowTCBList();

	payload_data dsrdata;

	memcpy(dsrdata, (uint8_t*) &tcp_pckt, sizeof(tcp_pckt));

	//send TCP segment using layer 3 (DSR) primitive
	return SendDSR(node, dsrdata);
}


//Function to restart Close Connection  procedure for a given destination node
void Restart_Close_Connection (void *datap){

	uint8_t dstnode = (uint8_t) datap;

	incTCBConnAttempt(dstnode);

	//check if maximum number of retransmissions reached
	if (getTCBConnAttempt(dstnode) > ConnClosLimit){
		printf("Maximum re-sending times reached. Terminating connection to dst node %d...\n",dstnode);

		//Liberate resources for destination node
		CloseSenderConnection(dstnode);
		
	}else{
		//Restart Close Connection to given destination node
		if (Init_Close_Connection(dstnode)) printf("Closing connection attempt %d to %d\n",getTCBConnAttempt(dstnode),dstnode);
		
		//Call Restart Close Connection Timer
		Timers_add(ConnClos_timeout*1024, Restart_Close_Connection, dstnode);
	}
}

//TCP call function to close a connection to a given target node
void CloseTCP(netaddr_t dstnode){

	if (!isConnectionInTCBList(dstnode)){
		printf("ERROR: no existent connection to node: %d\n",dstnode);
	}
	else{
		
		if (DEBUGTCP) ShowTCBList();
		if (Init_Close_Connection(dstnode)) {
			printf("Closing connection to %d...\n",dstnode);

			//increment SEQ number for the connection
			updateTCBSEQ(dstnode);

			//update TCB entry to FINWAIT1 state
			setTCBState(dstnode, FINWAIT1);
		}

		//initialize Close Connection Timer
		Timers_add(ConnClos_timeout*1024, Restart_Close_Connection, dstnode);
	}
}

//Function to send a termination response back to the initiator in Close Connection procedure
bool SendTCPFINACK(netaddr_t dstnode){

	tcp_packet_t tcp_pckt;

	//encapsulate TCP segment header in data field before sending it
	Init_TCP_Pckt(&tcp_pckt);
	tcp_pckt.SEQ_number = getTCBSEQ(dstnode);
	
	//set FIN and ACK control bits
	tcp_pckt.ACK_number = getTCBACK(dstnode);
	tcp_pckt.control = FINACK_CB;

	if (DEBUGTCP) ShowTCBList();

	payload_data dsrdata;
	memcpy(dsrdata, (uint8_t*) &tcp_pckt, sizeof(tcp_pckt));

	//send TCP segment using layer 3 (DSR) primitive
	return SendDSR(dstnode, dsrdata);
}

//Function to close connection at the initiator after termination timeout
void CloseSenderConnection(void *datap){

	uint8_t dstnode = (uint8_t) datap;

	//Remove all active timers to the given target node
	RemoveActiveTimers(dstnode);
	//Erase all Send Buffer entries to given target node
	ClearSendBufferEntries(dstnode);
	//Erase TCB entry for destination node
	DeleteTCBEntry(dstnode);
	
	P5OUT ^= 0x80;
	if (DEBUGTCP) printf("Connection resources liberated to %d\n",dstnode);
	printf("CONNECTION CLOSED to %d\n",dstnode);
	ShowTCPTestSender();

}

 //Function to prompt test metric results recorded by initiator node
void ShowTCPTestSender(){

	printf("\nINITIATOR NODE TEST RESULTS:\n\n");
	printf("TCP data segments sent: %d\n",SentTCPDataCtr);
	printf("TCP segment retransmissions: %d\n",RetxTCPDataCtr);
	printf("TCP data segments dropped: %d\n",DropTCPDataCtr);
	printf("DSR packets sent: %d\n",SentDataPcktCtr);

}

//Function to prompt test metric results recorded by receiver node
void ShowTCPTestReceiver(){

	printf("\nRECEIVER NODE TEST RESULTS:\n\n");
	printf("TCP data segments received: %d\n",RcvTCPDataCtr);
	//printf("TCP segments acknowledged: %d\n",AckTCPDataCtr);
	printf("DSR packets received: %d\n\n",RcvDataPcktCtr);

}

//Function to clear all Send Buffer entries to a given target node
void ClearSendBufferEntries(netaddr_t dstnode){

	int i;

	//traverse Send Buffer to look for the corresponding entry
	for(i=0; i<TAM_BUFF; i++){
		if (compAddr(SendBuffer[i].dst, dstnode)){
			SendBuffer[i].dst = 0;
			SendBuffer[i].inicount = 0;
			SendBuffer[i].retxcount = 0;
			SendBuffer[i].sgmnmbr = 0;
			Clear_TCP_Payload(SendBuffer[i].data);
		}
	}
}

//Function to remove all active timers to a given target node
void RemoveActiveTimers(netaddr_t dstnode){

	int i;
	//traverse Send Buffer to look for the corresponding entry
	for(i=0; i<TAM_BUFF; i++){
		if (compAddr(SendBuffer[i].dst, dstnode)){
			//remove timer using first Send Buffer entry
			Timers_remove(Restart_SendTCP, &SendBuffer[i]);
		}
	}
}

